@@ -1,3 +1,5 @@
+from typing import Iterable
+
 from pyrsistent import pvector, pmap
 import colander as col
 
@@ -9,17 +11,21 @@ class SchemaNode(col.SchemaNode):
     """ Colander's SchemaNode doesn't show node type in it's repr,
     we fix it with this subclass.
     """
+    children: Iterable
+
     def __repr__(self) -> str:
         return f'SchemaNode({self.typ})'
 
 
-class DictSchema(SchemaNode):
-    @staticmethod
-    def schema_type():
-        return col.Mapping(unknown='preserve')
+class TupleSchema(col.TupleSchema):
+    pass
 
 
-class SetSchema(col.SequenceSchema):
+class SequenceSchema(col.SequenceSchema):
+    pass
+
+
+class SetSchema(SequenceSchema):
     def __init__(self, *args, frozen=False, **kwargs):
         super().__init__(*args, **kwargs)
         self.frozen = frozen
@@ -41,7 +47,7 @@ class PVectorSchema(col.SequenceSchema):
         return pvector(r)
 
 
-class PMapSchema(DictSchema):
+class PMapSchema(SchemaNode):
     def deserialize(self, *args, **kwargs):
         r = super().deserialize(*args, **kwargs)
         if r in (Null, None):
